#include "lidar_lite.h"
#include <exceptions/lidar_lite_exceptions.h>

#include <ctime.h>

//******************************************************************************
// PUBLIC API
//******************************************************************************

// The device has a 7-bit slave address with a default value of 0x62.
// Fill in new address here if changed.
// See operating manual for instructions.
unsigned char DEVICE_ID = 0x62;

CLidarLite::CLidarLite(const std::string& serial, const int config)
{
  this->serial_ = serial;
  this->config_ = config;
  this->status_ = IDLE;
  this->max_unbiases_meas_ = 0;

  // On power-up or reset, the device performs a self-test sequence and
  // initializes
  // all registers with default values. After roughly 22 ms distance
  // measurements
  // can be taken with the I2C interface or the Mode Control Pin.
  usleep(22000);
}

CLidarLite::~CLidarLite(void)
{
  if (this->status_ == RUNNING) {
    reset();
    close();
  }
}

void
CLidarLite::open(void)
{
  if (this->status_ != IDLE)
    throw CLidarLiteException(
      _HERE_,
      "[CLidarLite]: Device cannot be OPENED because it is already running.");

  try {
    this->adapter_ = new CUSBI2C("adapter");

    // open serial port
    std::cout << "[CLidarLite]: Opening Lidar Lite. " << std::endl;
    this->adapter_->open(this->serial_);
    usleep(5000); // Delay to initialize sensor.

    std::cout << "  - USB_I2C serial number: " << this->serial_ << std::endl;
    std::cout << "  - USB_I2C firmware revision: "
              << (int)this->adapter_->get_revision() << std::endl;

    // Set gpio 2 and 3 as i2c
    this->adapter_->config_gpio(gpio2, i2c);
    this->adapter_->config_gpio(gpio3, i2c);

  } catch (CException& e) {
    throw CLidarLiteException(_HERE_, e.what());
  }

  // Configure device
  config_dev();

  // Set status to running
  this->status_ = RUNNING;
}

void
CLidarLite::close(void)
{
  if (this->status_ != RUNNING)
    throw CLidarLiteException(
      _HERE_, "Device cannot be CLOSED because it is not running.");

  if (this->adapter_ != NULL) {
    this->adapter_ = NULL;
    delete this->adapter_;
  }
}

void
CLidarLite::reset(void)
{
  if (this->status_ != RUNNING)
    throw CLidarLiteException(
      _HERE_, "Device cannot be RESET because it is not running.");

  // reset operation
  write(ACQ_COMMAND, 0x00);
  std::cout << "[CLidarLite]: Device RESET done." << std::endl;

  // On power-up or reset, the device performs a self-test sequence and
  // initializes
  // all registers with default values. After roughly 22 ms distance
  // measurements
  // can be taken with the I2C interface or the Mode Control Pin.
  usleep(22000);

  // Configure device
  config_dev();
}

void
CLidarLite::set_new_config(const int config)
{
  this->config_ = config;

  // Configure device
  config_dev();
}

int
CLidarLite::get_range(bool biasCorrection)
{
  if (this->status_ != RUNNING)
    throw CLidarLiteException(_HERE_, "RANGE MEASUREMENT cannot be obtained "
                                      "because the device is not running.");

  // Initiate an aquisition.
  if (biasCorrection || this->max_unbiases_meas_ > 99) {
    if (this->max_unbiases_meas_ > 99)
      std::cout << "\033[1;33m"
                << "[CLidarLite]: - WARNING : "
                << "\033[1;37m\033[0m"
                << "Forcing bias correction. Maximum measurements reached "
                   "without bias correction (100)."
                << std::endl;
    write(ACQ_COMMAND, 0x04);
    this->max_unbiases_meas_ = 0;
  } else {
    write(ACQ_COMMAND, 0x03);
    ++this->max_unbiases_meas_;
  }

  // Read two bytes from register 0x8f and save
  int range = 0;
  range = read(FULL_DELAY_HIGH, 2);

  return range;
}

std::vector<int>
CLidarLite::get_correlation(const int numberOfReadings)
{
  if (this->status_ != RUNNING)
    throw CLidarLiteException(_HERE_, "CORRELATION between measurements cannot "
                                      "be obtained because the device is not "
                                      "running.");

  if (numberOfReadings < 2 || numberOfReadings > 1024)
    throw CLidarLiteException(
      _HERE_, "CORRELATION between measurements cannot be obtained. Incorrect "
              "selecetd NUMBER OF READINGS, which must be between [2~1024].");

  // Object to return
  std::vector<int> correlationValues;
  correlationValues.resize(numberOfReadings);

  //  Select memory bank
  write(ACQ_SETTINGS, 0xc0);

  // Test mode enable
  write(COMMAND, 0x07);

  for (int ii = 0; ii < numberOfReadings; ii++) {
    // Read two bytes from 0xd2
    unsigned char addr = addr_to_read_2bytes(
      CORR_DATA); // See header file for a description of this functionality.
    unsigned char data[2];
    try {
      this->adapter_->write(DEVICE_ID, &addr, 1);
      this->adapter_->read(DEVICE_ID, data, 2);
    } catch (CException& e) {
      throw CLidarLiteException(_HERE_, e.what());
    }

    // Low byte is the value of the correlation record
    correlationValues.at(ii) = (int)data[0];

    // High byte is the sign from the record. If upper byte lsb is set, the
    // value is negative.
    if ((int)data[1] == 1)
      correlationValues.at(ii) = -correlationValues.at(ii);
  }

  // test mode disable
  write(COMMAND, 0x00);

  return correlationValues;
}

//******************************************************************************
// PRIVATE: DEVICE SPECIFIC FUNCTIONS
//******************************************************************************

int
CLidarLite::read(unsigned char addr, int len)
{
  if (len != 1 && len != 2)
    throw CLidarLiteException(_HERE_, "Expected read data with invalid length. "
                                      "This function only allows to read 1 or "
                                      "2 bytes");

  if (len == 2)
    addr = addr_to_read_2bytes(
      addr); // See header file for a description of this functionality.

  // Read STATUS. Repeat until bit 0 (LSB) goes low.
  wait_busy();
  
  unsigned char data[len];
  try {
    this->adapter_->write(DEVICE_ID, &addr, 1);
    this->adapter_->read(DEVICE_ID, data, len);
  } catch (CException& e) {
    throw CLidarLiteException(_HERE_, e.what());
  }

  int val;
  if (len == 1)
    val = data[0];
  else if (len == 2) {
    unsigned short int read_val = data[0] * 256 + data[1];
    val = (int)read_val;
  }
  return val;
}

void
CLidarLite::wait_busy(void)
{
  try {
    unsigned char addr = STATUS;
    this->adapter_->write(DEVICE_ID, &addr, 1);
  } catch (CException& e) {
    throw CLidarLiteException(_HERE_, e.what());
  }

  bool busy = true;
  int busyCounter = 0;

  try {
    while (busy) {
      unsigned char data;
      this->adapter_->read(DEVICE_ID, &data, 1);
      busy = data & 0x01;
      busyCounter++; // Increment busyCounter for timeout
      // Handle timeout condition, exit while loop and goto bailout
      if (busyCounter > 99)
        break;
    }
  } catch (CException& e) {
    throw CLidarLiteException(_HERE_, e.what());
  }
}

unsigned short int
CLidarLite::addr_to_read_2bytes(unsigned char first_byte_addr)
{
  unsigned short int val = (0x01 << 7) | first_byte_addr;
  return val;
}

void
CLidarLite::write(unsigned char addr, unsigned char cmd)
{
  try {
    this->adapter_->write_reg(DEVICE_ID, addr, &cmd, 1);
  } catch (CException& e) {
    throw CLidarLiteException(_HERE_, e.what());
  }
}

void
CLidarLite::config_dev(void)
{
  if (this->config_ < 0 || this->config_ > 5)
    throw CLidarLiteException(_HERE_, "Device cannot be CONFIGURED. Wrong "
                                      "configuration mode. Please select a "
                                      "value between [0~5].");

  try {
    switch (this->config_) {
      case 0:                          // Default mode, balanced performance
        write(SIG_COUNT_VAL, 0x80);    // Default
        write(ACQ_CONFIG_REG, 0x08);   // Default
        write(THRESHOLD_BYPASS, 0x00); // Default
        break;

      case 1: // Short range, high speed
        write(SIG_COUNT_VAL, 0x1d);
        write(ACQ_CONFIG_REG, 0x08);   // Default
        write(THRESHOLD_BYPASS, 0x00); // Default
        break;

      case 2:                       // Default range, higher speed short range
        write(SIG_COUNT_VAL, 0x80); // Default
        write(ACQ_CONFIG_REG, 0x00);
        write(THRESHOLD_BYPASS, 0x00); // Default
        break;

      case 3: // Maximum range
        write(SIG_COUNT_VAL, 0xff);
        write(ACQ_CONFIG_REG, 0x08);   // Default
        write(THRESHOLD_BYPASS, 0x00); // Default
        break;

      case 4: // High sensitivity detection, high erroneous measurements
        write(SIG_COUNT_VAL, 0x80);  // Default
        write(ACQ_CONFIG_REG, 0x08); // Default
        write(THRESHOLD_BYPASS, 0x80);
        break;

      case 5: // Low sensitivity detection, low erroneous measurements
        write(SIG_COUNT_VAL, 0x80);  // Default
        write(ACQ_CONFIG_REG, 0x08); // Default
        write(THRESHOLD_BYPASS, 0xb0);
        break;
    }
    std::cout << "[CLidarLite]: Configuration mode set to: " << this->config_
              << std::endl;
  } catch (CException& e) {
    throw CLidarLiteException(_HERE_, e.what());
  }
}

//******************************************************************************
// PRIVATE: MISC FUNCTIONS
//******************************************************************************

void
CLidarLite::blink_led_usb_adapter(void)
{
  if (this->status_ != RUNNING)
    throw CLidarLiteException(
      _HERE_,
      "USB_I2C adapter LED will not BLINK because Lidar-Lite is not running.");

  try {
    for (int ii = 0; ii < 2; ++ii) {
      this->adapter_->turn_led_off();
      usleep(200000);
      this->adapter_->turn_led_on();
      usleep(200000);
    }
  } catch (CException& e) {
    throw CLidarLiteException(_HERE_, e.what());
  }
}

std::string
CLidarLite::hex_to_str(const int& val)
{
  std::ostringstream ret;
  ret << "0x" << std::setfill('0') << std::setw(2) << std::hex << val;
  return ret.str();
}
